@@ -8,14 +8,17 @@ import scipy
 from scipy.sparse import csr_matrix
 from scipy.sparse import csc_matrix
 from pyrsb import rsb_matrix
-from pyrsb import _print_vec, rsb_time, _err_check, rsb_dtype, _dt2dt
+from pyrsb import _get_rsb_threads, _print_vec, rsb_time, _err_check, rsb_dtype, _dt2dt
 import pytest
 from pytest import raises as assert_raises
 from time import sleep
 
 
-rsb_dtypes = [numpy.float32, numpy.float64, numpy.complex64, numpy.complex128 ]
+rsb_real_dtypes = [ numpy.float32, numpy.float64 ]
+rsb_cplx_dtypes = [ numpy.complex64, numpy.complex128 ]
+rsb_dtypes = rsb_real_dtypes + rsb_cplx_dtypes
 prv_t = rsb_dtype
+max_idx_bpnz = 8
 
 
 def test__err_check_ok():
@@ -29,6 +32,10 @@ def test__err_check_err():
         _err_check(1,want_strict=True)
 
 
+def test__get_rsb_threads():
+    assert _get_rsb_threads() > 0
+
+
 def gen_tri_csr_larger():
     V = [11.,11.]
     J = [0,9]
@@ -50,11 +57,28 @@ def gen_tri(dtype=prv_t):
     return [V,I,J,2,2,3]
 
 
+def gen_rect(dtype=prv_t):
+    V = numpy.array([11.0, 60.0],dtype=dtype)
+    I = [0, 5]
+    J = [0, 0]
+    return [V,I,J,6,1,2]
+
+
+@pytest.fixture(params=rsb_dtypes)
+def f_gen_rect(request):
+    return gen_rect(dtype=request.param)
+
+
 @pytest.fixture(params=rsb_dtypes)
 def f_gen_tri(request):
     return gen_tri(dtype=request.param)
 
 
+@pytest.fixture(params=rsb_cplx_dtypes)
+def f_gen_tri_complex(request):
+    return gen_tri(dtype=request.param)
+
+
 def gen_x(n,nrhs=1,order='C', dtype=prv_t):
     x = numpy.empty([n, nrhs], dtype=dtype, order=order)
     for i in range(nrhs):
@@ -112,6 +136,11 @@ def test_init_from_none_none():
     assert mat.shape == (0, 0)
     assert mat.nnz == 0
     assert mat._is_unsymmetric() == True
+    assert mat._get_symchar() == 'G'
+    #TODO; enable these one 1.2.0.10 and 1.3 available:
+    #assert mat._idx_bpnz() > 0
+    #assert mat._idx_bpnz() <= max_idx_bpnz
+    assert mat._total_size > 0
 
 
 def test_init_tuple_csr_f32():
@@ -120,6 +149,7 @@ def test_init_tuple_csr_f32():
     assert mat.nnz == nnz
     assert mat.shape == (nr, nc)
     assert mat._is_unsymmetric() == True
+    assert mat._get_symchar() == 'G'
 
 
 def test_init_tuple_csr():
@@ -128,6 +158,7 @@ def test_init_tuple_csr():
     assert mat.nnz == nnz
     assert mat.shape == (nr, nc)
     assert mat._is_unsymmetric() == True
+    assert mat._get_symchar() == 'G'
 
 
 def test_init_tuple_csr_larger():
@@ -162,6 +193,9 @@ def test_init_tuples(f_gen_tri):
     assert mat.shape == (nr, nc)
     assert mat.nnz == nnz
     assert mat._is_unsymmetric() == True
+    assert mat._get_typechar() in [ 'S', 'D', 'C', 'Z' ]
+    assert mat._idx_bpnz() > 0
+    assert mat._idx_bpnz() <= max_idx_bpnz
 
 
 def test_init_tuples_and_dims(f_gen_tri):
@@ -170,6 +204,8 @@ def test_init_tuples_and_dims(f_gen_tri):
     assert mat.shape == (nr, nc)
     assert mat.nnz == nnz
     assert mat._is_unsymmetric() == True
+    assert mat._idx_bpnz() > 0
+    assert mat._idx_bpnz() <= max_idx_bpnz
 
 
 def test_init_tuples__raises(f_gen_tri):
@@ -247,6 +283,16 @@ def test_nonzero(f_gen_tri):
     assert ( cJ == rJ ).all()
 
 
+def test_render_to_stdout():
+    mat = rsb_matrix([1,1],dtype='d')
+    mat.render()
+
+
+# def test_render_to_stdout():
+#     mat = rsb_matrix([1,1],dtype='d')
+#     mat.render("one.eps")
+
+
 def test_io_bytes_ctor(f_gen_tri):
     for dtype in rsb_dtypes:
         [sV,sI,sJ,nr,nc,nnz] = f_gen_tri
@@ -345,12 +391,53 @@ def test__find_block(f_gen_tri):
     assert ( J == rJ ).all()
 
 
+def test__otn2obc_rect_n(f_gen_rect):
+    [V,I,J,nr,nc,nnz] = f_gen_rect
+    rmat = rsb_matrix((V, (I, J)),[nr,nc])
+    nrhs = 2
+    (cm_o,cm_ldB,cm_ldC) = rmat._otn2obc(False,'N',nrhs)
+    assert ( (cm_ldB,cm_ldC) == ( nc, nr ) )
+    (rm_o,rm_ldB,rm_ldC) = rmat._otn2obc(True ,'N',nrhs)
+    assert ( (rm_ldB,rm_ldC) == ( nrhs, nrhs ) )
+
+
+def test__otn2obc_rect_t(f_gen_rect):
+    [V,I,J,nr,nc,nnz] = f_gen_rect
+    rmat = rsb_matrix((V, (I, J)),[nr,nc])
+    nrhs = 2
+    (cm_o,cm_ldB,cm_ldC) = rmat._otn2obc(False,'T',nrhs)
+    assert ( (cm_ldB,cm_ldC) == ( nr, nc ) )
+    (rm_o,rm_ldB,rm_ldC) = rmat._otn2obc(True ,'T',nrhs)
+    assert ( (rm_ldB,rm_ldC) == ( nrhs, nrhs ) )
+
+
+def test__otn2obc_tri(f_gen_tri):
+    [V,I,J,nr,nc,nnz] = f_gen_tri
+    rmat = rsb_matrix((V, (I, J)),[nr,nc])
+    nrhs = 1
+    (cm_o,cm_ldB,cm_ldC) = rmat._otn2obc(False,'N',nrhs)
+    assert ( (cm_ldB,cm_ldC) == ( nr, nc ) )
+    (rm_o,rm_ldB,rm_ldC) = rmat._otn2obc(True ,'N',nrhs)
+    assert ( (rm_ldB,rm_ldC) == ( nrhs, nrhs ) )
+
+
 def test_init_tuples_sym(f_gen_tri):
     [V,I,J,nr,nc,nnz] = f_gen_tri
     mat = rsb_matrix((V, (I, J)),sym="S")
     assert mat.shape == (nr, nc)
     assert mat.nnz == nnz
     assert mat._is_unsymmetric() == False
+    assert mat._get_symchar() == 'S'
+
+
+def test_init_tuples_herm(f_gen_tri_complex):
+    [V,I,J,nr,nc,nnz] = f_gen_tri_complex
+    mat = rsb_matrix((V, (I, J)),sym="H",dtype=V.dtype) # Note: dtype not inherited from V.
+    assert mat.shape == (nr, nc)
+    assert mat.nnz == nnz
+    assert mat._is_unsymmetric() == False
+    assert mat._get_symchar() == 'H'
+    assert mat._is_complex()
 
 
 def test_init_tuples_wrong_sym(f_gen_tri):
@@ -365,6 +452,23 @@ def test_spmv__mul__(f_gen_mats):
     assert ( (rmat * x) == (cmat * x) ).all()
 
 
+def test_spmv_matvec(f_gen_mats):
+    [rmat,cmat] = f_gen_mats
+    x = gen_x(rmat.nc())
+    assert ( (rmat.matvec(x) ) == (cmat * x) ).all()
+
+
+def test_spmv_matvec_gmres():
+     for dtype in [float]:
+        A = csr_matrix([[3, 2, 0], [1, -1, 0], [0, 5, 1]], dtype=dtype)
+        A = rsb_matrix(A)
+        n = A.shape[0]
+        b = numpy.sin(numpy.array(range(1,n+1)))
+        from scipy.sparse.linalg import gmres
+        x, exitCode = gmres(A, b)
+        assert ( exitCode == 0 )
+
+
 def test_spmv_1D_N(f_gen_mats):
     [rmat,cmat] = f_gen_mats
     nrhs = 1
@@ -391,6 +495,13 @@ def test_spmv_1D_N_alpha(f_gen_mats):
         assert ( y == 2 * (cmat * x) ).all()
 
 
+def test_autotune_simple(f_gen_tri):
+    [V,I,J,nr,nc,nnz] = f_gen_tri
+    omat = rsb_matrix((V, (I, J)))
+    cmat = omat.copy()
+    assert( omat.todense() == cmat.todense() ).all()
+
+
 def test_autotune_simple(f_gen_tri):
     [V,I,J,nr,nc,nnz] = f_gen_tri
     omat = rsb_matrix((V, (I, J)))
@@ -459,7 +570,7 @@ def test_spmm_C_T(f_gen_mats):
     x = gen_x(rmat.nc(),nrhs,order)
     y = numpy.empty([rmat.nr(), nrhs], dtype=prv_t, order=order)
     y[:, :] = 0.0
-    rmat._spmm(x,y,transA=b'T')
+    rmat._spmm(x,y,transA='T')
     assert ( y == (cmat.transpose() * x) ).all()
 
 
@@ -498,7 +609,7 @@ def test_spmm_F_T(f_gen_mats):
     order = 'F'
     x = gen_x(rmat.nr(),nrhs,order)
     y = numpy.zeros([rmat.nc(), nrhs], dtype=prv_t, order=order)
-    rmat._spmm(x,y,transA=b'T')
+    rmat._spmm(x,y,transA='T')
     assert ( y == (cmat.transpose() * x) ).all()
 
 
